/* Functions for building and processing key-value lists */
/* in order to update a master list through a list of transactions */

/* Author: Zach Tomaszewski */
/* Date:   03 May 2003      */

#include "updlist.hpp"

/*
 Given two valid file stream, 
 Builds the necessary lists and runs the transactions 
   against the master (as per the other runTransactions).
   Files should contain one record per line. The ints
   give the length of each section of the line: transaction type 
   and key. Value extends to the end of the line. 
 Returns an updated master.
*/
KEYEDLIST runTransactions(ifstream& masterfile, ifstream& transfile,
                          int print = 0,
                          int trans = TRANSLEN, int key = KEYLEN){
  
   KEYEDLIST masterlist = fileToKeyedList(masterfile, trans, key); 
   KEYEDLIST translist = fileToKeyedList(transfile, trans, key);
   return runTransactions(masterlist, translist, print);
}


/*
 Takes a input stream.
   Files should contain one record per line. The ints
   give the length of each section of the line: transaction type 
   and key. Value extends to the end of the line. 
 Returns the resulting list. 
*/
KEYEDLIST fileToKeyedList(ifstream& file, 
                          int trans = TRANSLEN, int key = KEYLEN){
  string line, tstr, kstr, vstr;
  KEYEDLIST list;
  while(!file.eof()){
    getline(file, line);
    if (line.length() > 0) {
      tstr = line.substr(0, trans);
      kstr = line.substr(trans, key);
      vstr = line.substr(trans + key);
      list.push_back(KeyedItem(tstr, kstr, vstr));
    }
  }
  return list;
}


/*
 Given a master list and a transaction list,
  Adds elements to master if item.trans is INSERT
  Removes elements from from master if item.trans is a DELETE.
  If print is 0:
    works silently
  If print is 1:
    Prints error message on unfound deletes and repeat inserts.
  If print is 2:
    Also prints summary of work to the screen.
  If print is 3:
    Also prints resulting master.
 Returns the new master list.
*/
KEYEDLIST runTransactions(KEYEDLIST master, KEYEDLIST trans, 
                          int print = 0){
  int inserts = 0;
  int deletes = 0;
  int oldMasterLen = master.size();

  //add high values
  master.push_back(KeyedItem("  ", HIGHVALUE, ""));
  trans.push_back(KeyedItem("  ", HIGHVALUE, ""));

  KEYEDLIST::iterator masterIter = master.begin();
  KEYEDLIST::iterator transIter = trans.begin();

  while((*masterIter).key < HIGHVALUE || 
        (*transIter).key < HIGHVALUE) {
    if((*masterIter).key < (*transIter).key) {
       //master is less than trans.
       //Move ahead in master until find another transaction to do
       masterIter++;
    }else if((*masterIter).key > (*transIter).key) {
       //master is greater than trans
       //therefore have just passed where the current trans
       //needs to happen
       if ((*transIter).trans == DELETE){ //we didn't find the record
         if (print){
           cout << "Could not delete (does not exist in master): ";
           cout << (*transIter).key << endl;
         }
       }else if ((*transIter).trans == INSERT){ //needs to go here
         (*transIter).trans = CLEARTRANS;
         masterIter = master.insert(masterIter, *transIter);
         inserts++;
       }
       transIter++; //on to next transaction
    }
    else { // they must be equal
       if ((*transIter).trans == DELETE){ //we found it, so delete it
         masterIter = master.erase(masterIter);
         //erasing moves masterIter to next space
         deletes++;
       }else if ((*transIter).trans == INSERT){ //record already exists!
         if (print){
           cout << "Could not insert (already exists in master): "; 
           cout<< (*transIter).key << endl;
         }
       }
       transIter++; //on to next transaction
    }
  }

  //all done; remove master's high value
  master.pop_back();

  if (print >= 3) {
    printKeyedList(master);
  }
  if (print >= 2){
    cout << oldMasterLen << " entries in original master file." << endl;
    cout << inserts << " inserts performed." << endl;
    cout << deletes << " deletes performed." << endl;
    cout << master.size() << " entries in new master file." << endl;
  }
  return master;
}


/*
 Prints a KeyedList to the screen.
   Default is to the screen, but will print to a file if given a stream.
*/
void printKeyedList(KEYEDLIST list, ostream& out = cout){
  KEYEDLIST::iterator iter = list.begin();
  while(iter != list.end()){
    out << (*iter).trans << (*iter).key << (*iter).value << endl;
    iter++;
  }
}
